#!/usr/bin/env python
"""
Usage:
  mmi-runner <engine> <configfile> [-o <outputvar>...] [-g <globalvar>...] [--interval <interval>] [--disable-logger] [--pause] [--mpi <method>] [--track <server>] [--port <port>] [--bmi-class]
  mmi-runner -h | --help

Positional arguments:
  engine model  engine
  configfile    model configuration file

Optional arguments:
  -h, --help               show this help message and exit
  --interval <interval>    publish results every <interval> timesteps
  -o <outputvar>           output variables, will be broadcasted each <interval> timestep
  -g <globalvar>           global variables, will be send if requested
  --disable-logger         do not inject logger into the BMI library
  --pause                  start in paused mode, send update messages to progress
  --mpi <method>           communicate with mpi nodes using one of the methods: root (communicate with rank 0), all (one socket per rank)
  --port <port>            "random" or integer base port, port is computed as req/rep = port + rank*3 + 0, push/pull = port + rank*3 + 1, pub/sub = port + rank*3 + 2 [default: random]
  --track <tracker>        server to subscribe to for tracking
  --bmi-class              when used - the engine is assumed to be the full name of a Python class that implements bmi [default: bmi.wrapper.BMIWrapper]

"""
import os
import logging
import json
from six.moves.urllib.parse import urljoin
import itertools
import atexit
import platform

import docopt
import requests
import zmq
import zmq.eventloop.zmqstream
from zmq.eventloop import ioloop


import bmi.wrapper
from mmi import send_array, recv_array



logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

ioloop.install()

# see or an in memory numpy message:
# http://zeromq.github.io/pyzmq/serialization.html
def register(server, metadata):
    """register model at tracking server"""
    logger.debug("register at server %s with %s", server, metadata)
    # connect to server
    result = requests.post(urljoin(server, 'models'), data=json.dumps(metadata))
    logger.debug("got back %s", result.json())
    metadata["tracker"] = result.json()


def unregister(server, metadata):
    """unregister model at tracking server"""
    uuid = metadata["tracker"]["uuid"]
    # connect to server
    result = requests.delete(urljoin(server, 'models' + "/" + uuid))
    logger.debug("unregistered at server %s with %s: %s", server, uuid, result)


def process_incoming(model, sockets, data):
    """
    process incoming messages

    data is a dict with several arrays
    """
    # Check for new messages
    if not sockets:
        return
    # unpack sockets
    poller = sockets['poller']
    rep = sockets['rep']
    # pull = sockets['pull']
    pub = sockets['pub']
    items = poller.poll(10)
    for sock, n in items:
        for i in range(n):

            A, metadata = recv_array(sock)
            var = None

            # bmi actions
            if "update" in metadata:
                dt = float(metadata["update"])
                model.update(dt)
                metadata["dt"] = dt
            elif "get_var" in metadata:
                name = metadata["get_var"]
                # temporary implementation
                if metadata.get("copy", False):
                    var = model.get_var(name).copy()
                else:
                    var = model.get_var(name)
                if var is None:
                    logger.warning("Get_var returns None for %s" % name)
                metadata['name'] = name
                # assert socket is req socket
            elif "get_var_count" in metadata:
                # temporary implementation
                n = model.get_var_count()
                metadata['get_var_count'] = n
                # assert socket is req socket
            elif "get_var_rank" in metadata:
                # temporary implementation
                var_name = metadata['get_var_rank']
                n = model.get_var_rank(var_name)
                metadata['get_var_rank'] = n
                # assert socket is req socket
            elif "get_var_shape" in metadata:
                # temporary implementation
                var_name = metadata['get_var_shape']
                n = model.get_var_shape(var_name)
                metadata['get_var_shape'] = tuple([int(item) for item in n])
                # assert socket is req socket
            elif "get_var_type" in metadata:
                # temporary implementation
                var_name = metadata['get_var_type']
                n = model.get_var_type(var_name)
                metadata['get_var_type'] = n
                # assert socket is req socket
            elif "get_var_name" in metadata:
                i = int(metadata["get_var_name"])
                name = model.get_var_name(i)
                metadata['get_var_name'] = name
                # assert socket is req socket
            elif "set_var" in metadata:
                name = metadata["set_var"]
                # logger.debug("setting variable %s", name)
                model.set_var(name, A)
                metadata["name"] = name  # !?
            elif "set_var_slice" in metadata:
                name = metadata["set_var_slice"]
                logger.debug("setting variable %s", name)
                start = metadata["start"]
                count = metadata["count"]
                model.set_var_slice(name, start, count, A)
                metadata["name"] = name  # !?
            elif "set_var_index" in metadata:
                name = metadata["set_var_index"]
                logger.debug("setting variable %s using index index", name)
                index = metadata["index"]
                # TODO: test if this is fast enough.
                # Otherwise move to BMI ++ but that is
                # a bit of a burden on implementers
                var = model.get_var(name).copy()
                var.flat[index] = A
                model.set_var(name, var)
                metadata["name"] = name  # !?
            elif "get_current_time" in metadata:
                metadata["get_current_time"]
                t = model.get_current_time()
                metadata['get_current_time'] = t
            elif "get_time_step" in metadata:
                metadata["get_time_step"]
                dt = model.get_time_step()
                metadata['get_time_step'] = dt
            elif "get_end_time" in metadata:
                metadata["get_end_time"]
                t = model.get_end_time()
                metadata['get_end_time'] = t
            elif "get_start_time" in metadata:
                metadata["get_start_time"]
                t = model.get_start_time()
                metadata['get_start_time'] = t
                # assert socket is req socket
            # custom actions
            elif "remote" in metadata:
                assert metadata["remote"] in {
                    "play", "stop", "pause", "rewind", "quit"}
                model.state = metadata["remote"]
            elif "operator" in metadata:
                # TODO: support same operators as MPI_ops here....,
                # TODO: reduce before apply
                # TODO: assert pull socket
                pass
                # S = tuple(slice(*x) for x in action['slice'])
                # print(repr(arr[S]))
                # if action['operator'] == 'setitem':
                #     arr[S] = data
                # elif action['operator'] == 'add':
                #     arr[S] += data
            elif "initialize" in metadata:
                config_file = metadata["initialize"]
                model.initialize(str(config_file))
            elif "finalize" in metadata:
                model.finalize()
            else:
                logger.warn("got unknown message {} from socket {}".format(str(metadata), sock))
            if sock.socket_type == zmq.REP:
                # reply
                send_array(rep, var, metadata=metadata)
            # any getter requested through the pull socket?
            elif any(x.startswith("get_") for x in metadata) and sock.socket_type == zmq.PULL:
                # return through the pub socket
                send_array(pub, var, metadata=metadata)


def create_sockets(ports):
    context = zmq.Context()
    poller = zmq.Poller()

    # Socket to handle init data
    rep = context.socket(zmq.REP)
    # this was inconsequent: here REQ is for the client, we reply with REP.
    # PULL and PUB is seen from here, not from the client.
    # Is now renamed to PUSH and SUB: everything is seen from outside.
    if "REQ" in ports:
        rep.bind(
            "tcp://*:{port}".format(port=ports["REQ"])
        )
    else:
        ports["REQ"] = rep.bind_to_random_port(
            "tcp://*"
        )

    pull = context.socket(zmq.PULL)
    if "PUSH" in ports:
        pull.bind(
            "tcp://*:{port}".format(port=ports["PUSH"])
        )
    else:
        ports["PUSH"] = pull.bind_to_random_port(
            "tcp://*"
        )

    # for sending model messages
    pub = context.socket(zmq.PUB)
    if "SUB" in ports:
        pub.bind(
            "tcp://*:{port}".format(port=ports["SUB"])
        )
    else:
        ports["SUB"] = pub.bind_to_random_port(
            "tcp://*"
        )

    poller.register(rep, zmq.POLLIN)
    poller.register(pull, zmq.POLLIN)
    sockets = dict(
        poller=poller,
        rep=rep,
        pull=pull,
        pub=pub
    )
    return sockets


def runner(arguments, wrapper_kwargs={}, extra_metadata={}):
    """
    MMI Runner
    """
    # make a socket that replies to message with the grid

    # if we are running mpi we want to know the rank
    if arguments['--mpi']:
        import mpi4py.MPI
        comm = mpi4py.MPI.COMM_WORLD
        rank = comm.Get_rank()
        size = comm.Get_size()
    else:
        # or we assume it's 0
        rank = 0
        size = 1

    if arguments["--port"] == "random":
        # ports will be filled in
        ports = {}
    else:
        port = int(arguments['--port'])
        ports = {
            "REQ": port + 0,
            "PUSH": port + 1,
            "SUB": port + 2
        }

    # if we want to communicate with separate domains
    # we have to setup a socket for each of them
    if arguments['--mpi'] == 'all':
        # use a socket for each rank rank
        for port in ports:
            ports[port] += (rank * 3)

    # now we can create the zmq sockets and poller
    sockets = {}
    if rank == 0 or arguments['--mpi'] == 'all':
        # Create sockets
        sockets = create_sockets(ports)
        logger.debug("ZMQ MMI Ports:")
        for key, value in ports.items():
            logger.debug("%s = %d" % (key, value))

    # not so verbose
    # bmi.wrapper.logger.setLevel(logging.WARN)

    if arguments.get('--bmi-class'):
        full_name = arguments['<engine>']
        s = full_name.split('.')
        class_name = s[-1]
        module_name = full_name[:-len(class_name)-1]
        import importlib
        wrapper_module = importlib.import_module(module_name)
        wrapper_class = getattr(wrapper_module, class_name)
        model = wrapper_class()

    else:
        wrapper_class = bmi.wrapper.BMIWrapper
        model = wrapper_class(
            engine=arguments['<engine>'],
            configfile=arguments['<configfile>'],
            **wrapper_kwargs)

    # for replying to grid requests
    model.state = "play"
    if arguments["--pause"]:
        model.state = "pause"
        logger.info("model initialized and started in pause mode, waiting for requests ...")
    else:
        logger.info("model started and initialized, running ...")

    if arguments["--track"]:
        server = arguments["--track"]

        metadata = {}
        # update connection information from external service
        # You might want to disable this if you have some sort of sense of privacy
        try:
            metadata["ifconfig"] = requests.get("http://ipinfo.io/json").json()
        except:
            pass
        # except requests.exceptions.ConnectionError:
        #     logger.exception("Could not read ip info from ipinfo.io")
        #     pass
        # except simplejson.scanner.JSONDecodeError:
        #     logger.exception("Could not parse ip info from ipinfo.io")
        #     pass
        # node
        metadata["node"] = platform.node()
        metadata.update({
            "engine": arguments['<engine>'],
            "configfile": arguments['<configfile>'],
            "ports": ports,
            "rank": rank,
            "size": size
        })

        metadata_filename = os.path.join(
            os.path.dirname(arguments['<configfile>']),
            "metadata.json"
        )
        if os.path.isfile(metadata_filename):
            metadata.update(json.load(open(metadata_filename)))
        metadata.update(extra_metadata)

        register(server, metadata)

    if arguments["--track"]:
        atexit.register(unregister, server, metadata)
    # Start a reply process in the background, with variables available
    # after initialization, sent all at once as py_obj
    data = {
        var: model.get_var(var)
        for var
        in arguments['-g']
    }
    process_incoming(model, sockets, data)

    # Keep on counting indefinitely
    counter = itertools.count()
    logger.info("Entering timeloop...")
    for i in counter:
        while model.state == "pause":
            # keep waiting for messages when paused
            process_incoming(model, sockets, data)
        else:
            # otherwise process messages once and continue
            process_incoming(model, sockets, data)
        if model.state == "quit":
            break

        # paused ...
        model.update(-1)

        # check counter
        if arguments.get('--interval') and (i % int(arguments['--interval'])):
            continue

        for key in arguments['-o']:
            value = model.get_var(key)
            metadata = {'name': key, 'iteration': i}
            # 4ms for 1M doubles
            logger.debug("sending {}".format(metadata))
            if 'pub' in sockets:
                send_array(sockets['pub'], value, metadata=metadata)

    logger.info("Finalizing...")
    model.finalize()

def main():
    arguments = docopt.docopt(__doc__)
    runner(arguments)


if __name__ == '__main__':
    logging.basicConfig()

    logger.info("mmi-runner")
    main()



"""
Model Message Interface
"""

import datetime
import sys

import numpy as np
import zmq


__version__ = '0.1.11'


if sys.version_info > (3, ):
    buffer = memoryview


class NoResponseException(Exception):
    pass


class EmptyResponseException(Exception):
    pass


def send_array(
        socket, A=None, metadata=None, flags=0,
        copy=False, track=False, compress=None,
        chunksize=50 * 1000 * 1000
):
    """send a numpy array with metadata over zmq

    message is mostly multipart:
    metadata | array part 1 | array part 2, etc

    only metadata:
    metadata

    the chunksize roughly determines the size of the parts being sent
    if the chunksize is too big, you get an error like:
        zmq.error.Again: Resource temporarily unavailable
    """

    # create a metadata dictionary for the message
    md = {}
    # always add a timestamp
    md['timestamp'] = datetime.datetime.now().isoformat()

    # copy extra metadata
    if metadata:
        md.update(metadata)

    # if we don't have an array
    if A is None:
        # send only json
        md['parts'] = 0
        socket.send_json(md, flags)
        # and we're done
        return

    # add array metadata
    md['dtype'] = str(A.dtype)
    md['shape'] = A.shape
    # determine number of parts
    md['parts'] = int(np.prod(A.shape) // chunksize + 1)
    try:
        # If an array has a fill value assume it's an array with missings
        # store the fill_Value in the metadata and fill the array before sending.
        # asscalar should work for scalar, 0d array or nd array of size 1
        md['fill_value'] = np.asscalar(A.fill_value)
        A = A.filled()
    except AttributeError:
        # no masked array, nothing to do
        pass

    # send json, followed by array (in x parts)
    socket.send_json(md, flags | zmq.SNDMORE)

    # although the check is not strictly necessary, we try to maintain fast
    # pointer transfer when there is only 1 part
    if md['parts'] == 1:
        msg = memoryview(np.ascontiguousarray(A))
        socket.send(msg, flags, copy=copy, track=track)
    else:
        # split array at first dimension and send parts
        for i, a in enumerate(np.array_split(A, md['parts'])):
            # Make a copy if required and pass along the memoryview
            msg = memoryview(np.ascontiguousarray(a))
            flags_ = flags
            if i != md['parts'] - 1:
                flags_ |= zmq.SNDMORE
            socket.send(msg, flags_, copy=copy, track=track)
    return


def recv_array(
    socket, flags=0, copy=False, track=False, poll=None, poll_timeout=10000):
    """recv a metadata and an optional numpy array from a zmq socket

    Optionally provide poll object to use recv_array with timeout

    poll_timeout is in millis
    """
    if poll is None:
        md = socket.recv_json(flags=flags)
    else:
        # one-try "Lazy Pirate" method: http://zguide.zeromq.org/php:chapter4
        socks = dict(poll.poll(poll_timeout))
        if socks.get(socket) == zmq.POLLIN:
            reply = socket.recv_json(flags=flags)
            # note that reply can be an empty array
            md = reply
        else:
            raise NoResponseException(
                "Recv_array got no response within timeout (1)")

    if md['parts'] == 0:
        # No array expected
        A = None
    elif md['parts'] == 1:
        # although the check is not strictly necessary, we try to maintain fast
        # pointer transfer when there is only 1 part

        if poll is None:
            msg = socket.recv(flags=flags, copy=copy, track=track)
        else:
            # one-try "Lazy Pirate" method: http://zguide.zeromq.org/php:chapter4
            socks = dict(poll.poll(poll_timeout))
            if socks.get(socket) == zmq.POLLIN:
                reply = socket.recv(flags=flags, copy=copy, track=track)
                # note that reply can be an empty array
                msg = reply
            else:
                raise NoResponseException(
                    "Recv_array got no response within timeout (2)")
        buf = buffer(msg)
        A = np.frombuffer(buf, dtype=md['dtype'])
        A = A.reshape(md['shape'])

        if 'fill_value' in md:
            A = np.ma.masked_equal(A, md['fill_value'])

    else:
        # multi part array
        A = np.zeros(np.prod(md['shape']), dtype=md['dtype'])
        arr_position = 0
        for i in range(md['parts']):
            if poll is None:
                msg = socket.recv(flags=flags, copy=copy, track=track)
            else:
                # one-try "Lazy Pirate" method: http://zguide.zeromq.org/php:chapter4
                socks = dict(poll.poll(poll_timeout))
                if socks.get(socket) == zmq.POLLIN:
                    reply = socket.recv(flags=flags, copy=copy, track=track)
                    if not reply:
                        raise EmptyResponseException(
                            "Recv_array got an empty response (2)")
                    msg = reply
                else:
                    raise NoResponseException(
                        "Recv_array got no response within timeout (2)")
            buf = buffer(msg)
            a = np.frombuffer(buf, dtype=md['dtype'])
            A[arr_position:arr_position + a.shape[0]] = a[:]
            arr_position += a.shape[0]
        A = A.reshape(md['shape'])

        if 'fill_value' in md:
            A = np.ma.masked_equal(A, md['fill_value'])

    return A, md



import uuid
import json
# import itertools
# from threading import Thread
import logging

logging.basicConfig(level=logging.DEBUG)

logger = logging.getLogger(__name__)

import numpy as np
import six
import zmq
import osgeo.osr
import shapely.geometry

from zmq.eventloop.zmqstream import ZMQStream
from zmq.eventloop import ioloop
ioloop.install()

SOCKET_NAMES = {
    int(getattr(zmq, name)): name
    for name
    in {"PUSH", "PULL", "SUB",
        "PUB", "REQ", "REP", "PAIR"}
}


import tornado.websocket
import tornado.web
import tornado.ioloop

from . import send_array, recv_array


# TODO: implement somewhere else
class Views(object):
    @staticmethod
    def grid(context):
        meta = context["value"]
        # Get connection info
        node = meta['node']
        node = 'localhost'
        req_port = meta["ports"]["REQ"]
        ctx = context["ctx"]
        req = ctx.socket(zmq.REQ)
        req.connect("tcp://%s:%s" % (node, req_port))
        # Get grid variables
        send_array(req, A=None, metadata={"get_var": "xk"})
        xk, A = recv_array(req)
        send_array(req, A=None, metadata={"get_var": "yk"})
        yk, A = recv_array(req)
        send_array(req, A=None, metadata={"get_var": "flowelemnode"})
        yk, A = recv_array(req)
        # Spatial transform
        points = np.c_[xk, yk]
        logger.info("points shape: %s, values: %s", points.shape, points)
        src_srs = osgeo.osr.SpatialReference()
        src_srs.ImportFromEPSG(meta["epsg"])
        dst_srs = osgeo.osr.SpatialReference()
        dst_srs.ImportFromEPSG(4326)

        transform = osgeo.osr.CoordinateTransformation(src_srs, dst_srs)
        wkt_points = transform.TransformPoints(points[:1000])
        geom = shapely.geometry.MultiPoint(wkt_points)

        geojson = shapely.geometry.mapping(geom)
        return geojson


views = Views()


class WebSocket(tornado.websocket.WebSocketHandler):
    def __init__(self, application, request, **kwargs):
        # self.zmqstream = kwargs.pop('zmqstream')
        tornado.websocket.WebSocketHandler.__init__(
            self, application, request, **kwargs)
        self.metadata = None

    def initialize(self, database, ctx):
        self.database = database
        self.ctx = ctx

    def open(self, key):
        logger.debug("websocket opened for key %s", key)
        self.key = key
        # openeing corresponding socket of model

        # open push socket to forward incoming zmq messages
        push = self.ctx.socket(zmq.PUSH)
        pull_port = self.database[key]["ports"]['PULL']
        node = "localhost"
        # node = self.database[key]["node"]

        push.connect("tcp://%s:%d" % (node, pull_port))
        self.pushstream = ZMQStream(push)

        sub = self.ctx.socket(zmq.SUB)
        pub_port = self.database[key]["ports"]['PUB']
        sub.connect("tcp://%s:%d" % (node, pub_port))
        # Accept all messages
        sub.setsockopt(zmq.SUBSCRIBE, '')
        self.substream = ZMQStream(sub)

        def send(messages):
            """forward messages to this websocket"""
            logger.info("received %s messages", len(messages))
            for message in messages:
                try:
                    json.loads(message)
                    binary = False
                except ValueError:
                    binary = True
                self.write_message(message, binary)
        self.substream.on_recv(send)

    def on_message(self, message):
        # unicode, metadata message
        logger.debug("got message %20s, text: %s", message, isinstance(message, six.text_type))

        # Let's try and forward it.
        # use the zmqstream as a socket (send, send_json)
        socket = self.pushstream

        if isinstance(message, six.text_type):
            metadata = json.loads(message)
            logger.debug("got metadata %s", metadata)
            if not ("dtype" in metadata):
                # no array expected, pass along:
                A = None
                logger.debug("sending metadata %s to %s of type %s", metadata, socket, SOCKET_NAMES[socket.socket.type])
                send_array(socket, None, metadata)
                self.metadata = None
            else:
                # We expect another message
                self.metadata = metadata
        else:
            # assume we already have metadata
            if not self.metadata:
                logger.warn("got message without preceding metadata")
            dtype = self.metadata['dtype']
            shape = self.metadata['shape']
            # unpack array
            A = np.fromstring(message, dtype=dtype)
            A = A.reshape(shape)
            # pass it along to the socket (push)
            send_array(socket, A, self.metadata)
            self.metadata = None

    def on_close(self):
        logger.debug("websocket closed")

    def check_origin(self, origin):
        """connect from everywhere"""
        return True


class MainHandler(tornado.web.RequestHandler):

    def initialize(self, database):
        self.database = database

    def get(self):
        self.set_header("Content-Type", "application/json")
        self.write(json.dumps(self.database))


class ModelHandler(tornado.web.RequestHandler):

    def initialize(self, database, ctx=None):
        self.database = database
        self.ctx = ctx

    def get(self, key=None, view=None):
        """Register a new model (models)"""
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Content-Type", "application/json")
        if key is not None:
            value = {}
            value.update(self.database[key])
            if view is not None:
                # generate a context with the relevant variables
                context = {}
                context["value"] = value
                context["ctx"] = self.ctx
                result = json.dumps(getattr(views, view)(context))
            else:
                result = json.dumps(value)
        else:
            result = json.dumps(self.database.values())

        self.write(result)

    def post(self):
        """Register a new model (models)"""
        self.set_header("Content-Type", "application/json")
        key = uuid.uuid4().hex
        metadata = json.loads(self.request.body)
        metadata["uuid"] = key
        self.database[key] = metadata
        result = json.dumps({"uuid": key})
        self.write(result)

    def put(self, key, *args, **kwargs):
        # TODO: show a list of running models
        self.database[key] = json.loads(self.request.body)

    def delete(self, key, *args, **kwargs):
        del self.database[key]


def app():
    # register socket
    ctx = zmq.Context()
    # Use something more webscale when needed (like dbm)
    database = {}
    application = tornado.web.Application([
        (r"/", MainHandler, {"database": database}),
        # todo use an id scheme to attach to multiple models
        (r"/models/?", ModelHandler, {"database": database}),
        (r"/models/(\w+)/?", ModelHandler, {"database": database}),
        (r"/models/(?P<key>\w+)/(?P<view>\w*)/?", ModelHandler, {"database": database, "ctx": ctx}),
        (r"/mmi/(\w+)/?", WebSocket, {"database": database, "ctx": ctx}),
    ])
    # You reuse this app in other wsgi applications
    # Encapsulate it using a tornado.wsgi.WSGIAdapter
    # Won't work with websockets for now....
    # TODO some connection heartbeat that removes stale connections
    return application


def main():
    # Use common port not in services...
    # Normally you'd run this behind an nginx
    logger.info('mmi-tracker')
    application = app()
    application.listen(22222)
    logger.info('serving at port 22222')
    tornado.ioloop.IOLoop.instance().start()

if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Usage:
  mmi-curl <url> [--request=<command>] [--metadata=<json>]
  mmi-curl -h | --help

Positional arguments:
  url  connection string used for sending/receiving messages

Optional arguments:
  -h, --help               show this help message and exit
  -d --metadata=<json>     send a message with metadata [default: {}]
  -X --request=<command>   use the following method of connection (REQ|SUB|PUSH) [default: SUB]

"""
import logging
import json

import zmq
import zmq.eventloop.zmqstream
from zmq.eventloop import ioloop
import numpy as np
import docopt

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from mmi import send_array, recv_array

def main():
    arguments = docopt.docopt(__doc__)
    logger.info(arguments)
    context = zmq.Context()
    metadata = json.loads(arguments['--metadata'])
    url = arguments['<url>']
    if arguments['--request'] == 'SUB':
        sub = context.socket(zmq.SUB)
        sub.connect(url)
        sub.setsockopt(zmq.SUBSCRIBE, '')
        while True:
            arr, metadata = recv_array(sub)
            logger.info("metadata: %s\n array: %s", metadata, arr)
    elif arguments['--request'] == 'REQ':
        req = context.socket(zmq.REQ)
        req.connect(url)
        value = None
        send_array(req, value, metadata=metadata)
        # wait for reply
        arr, metadata = recv_array(req)
        logger.info("metadata: %s\n array: %s", metadata, arr)
    elif arguments['--request'] == 'PUSH':
        req = context.socket(zmq.PUSH)
        req.connect(url)
        value = None
        send_array(req, value, metadata=metadata)
        logger.info("metadata: %s", metadata)








from setuptools import setup, find_packages
import os
import re
import codecs
# get py3 compatible open
from io import open

here = os.path.abspath(os.path.dirname(__file__))

# Read the version number from a source file.
# Code taken from pip's setup.py
def find_version(*file_paths):
    # Open in Latin-1 so that we avoid encoding errors.
    # Use codecs.open for Python 2 compatibility
    with codecs.open(os.path.join(here, *file_paths), 'r', 'latin1') as f:
        version_file = f.read()

    # The version line must have the form
    # __version__ = 'ver'
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# Get the long description from the relevant file
with open('DESCRIPTION.rst', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="mmi",
    version=find_version('mmi', '__init__.py'),
    description="Model Message Interface",
    long_description=long_description,

    # The project URL.
    url='http://github.com/SiggyF/python-mmi',

    # Author details
    author='Fedor Baart',
    author_email='fedor.baart@deltares.nl',

    # Choose your license
    license='GPLv3+',

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        'Topic :: System :: Distributed Computing',

        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
    ],

    # What does your project relate to?
    keywords='array messages model',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages.
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    # If there are data files included in your packages, specify them here.
    package_data={
        # 'sample': ['*.dat'],
    },
    setup_requires=[
        'sphinx',
        'sphinx_rtd_theme'
    ],
    install_requires=[
        'numpy',
        'pyzmq',
        'docopt',
        'bmi-python',
        'tornado',
        'six',
        'shapely'

    ],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'mmi-runner=mmi.runner:main',
            'mmi-curl=mmi.curl:main',
            'mmi-tracker=mmi.tracker:main',
        ],
    },
)
